import 'package:flutter_test/flutter_test.dart';

import 'package:benchmark/benchmark.dart' as benchmark;

void main() {
  setUp(() => benchmark.reset());

  group('benchmark', () {
    test('The duration should be used when passed', () {
      final watch = Stopwatch();

      benchmark.setUp(() => watch.start());
      benchmark.tearDown(() => watch.stop());

      benchmark.benchmark(
        'benchmark',
        () {},
        duration: Duration(seconds: 3),
      );

      benchmark.measure();

      expect(watch.elapsedMilliseconds, greaterThanOrEqualTo(3000));
    });
  });

  group('setUp methods', () {
    test('The setUpAll should be called in the right order', () {
      var result = <String, int>{
        'firstSetUpAll': 0,
        'secondSetUpAll': 0,
      };

      benchmark.setUpAll(() => result['firstSetUpAll']++);

      benchmark.group('sub group', () {
        benchmark.setUpAll(() => result['secondSetUpAll']++);

        benchmark.benchmark('benchmark', () {}, iterations: 0);
      });

      benchmark.measure();

      expect(result['firstSetUpAll'], equals(1),
          reason: 'The first setUpAll should be called once');
      expect(result['secondSetUpAll'], equals(1),
          reason: 'The second setUpAll should be called once');
    });

    test('The setUp should be called before benchmarks, and for each benchmark',
        () {
      final result = <String, int>{
        'firstSetUp': 0,
        'secondSetUp': 0,
      };

      benchmark.setUp(() => result['firstSetUp']++);

      benchmark.group('sub group', () {
        benchmark.setUp(() => result['secondSetUp']++);

        benchmark.benchmark('benchmark', () {}, iterations: 0);
      });

      benchmark.measure();

      expect(result['firstSetUp'], equals(1),
          reason: 'The first setUp should be called once');
      expect(result['secondSetUp'], equals(1),
          reason: 'The second setUp should be called once');
    });

    //   test('The setUpEach should be called before iteration of a benchmark', () {
    //     final result = <String, int>{
    //       'firstSetUpEach': 0,
    //       'secondSetUpEach': 0,
    //     };

    //     benchmark.setUpEach(() => result['firstSetUpEach']++);

    //     benchmark.group('sub group', () {
    //       benchmark.setUpEach(() => result['secondSetUpEach']++);

    //       benchmark.benchmark(
    //         'benchmark',
    //         () {},
    //         iterations: 0,
    //         duration: Duration(microseconds: 0),
    //       );
    //     });

    //     benchmark.measure();

    //     expect(result['firstSetUpEach'], 714811,
    //         reason: 'The first setUpEach should be called 714811 times');
    //     expect(result['secondSetUpEach'], 714811,
    //         reason: 'The second setUpEach should be called 714811 times');
    //   });
  });

  group('tearDown methods', () {
    test('The tearDownAll should be called in the right order', () {
      var result = <String, int>{
        'firstTearDownAll': 0,
        'secondTearDownAll': 0,
      };

      benchmark.tearDownAll(() => result['firstTearDownAll']++);

      benchmark.group('sub group', () {
        benchmark.tearDownAll(() => result['secondTearDownAll']++);

        benchmark.benchmark('benchmark', () {}, iterations: 0);
      });

      benchmark.measure();

      expect(result['firstTearDownAll'], equals(1),
          reason: 'The first tearDownAll should be called once');
      expect(result['secondTearDownAll'], equals(1),
          reason: 'The second tearDownAll should be called once');
    });

    test(
        'The tearDown should be called after benchmarks, and for each benchmark',
        () {
      final result = <String, int>{
        'firstTearDown': 0,
        'secondTearDown': 0,
      };

      benchmark.tearDown(() => result['firstTearDown']++);

      benchmark.group('sub group', () {
        benchmark.tearDown(() => result['secondTearDown']++);

        benchmark.benchmark('benchmark', () {}, iterations: 0);
      });

      benchmark.measure();

      expect(result['firstTearDown'], equals(1),
          reason: 'The first tearDown should be called once');
      expect(result['secondTearDown'], equals(1),
          reason: 'The second tearDown should be called once');
    });
  });
}
